from pkgtools.pypi import PyPIXmlRpc, real_name
try:
    from urllib2 import HTTPError
except ImportError:
    from urllib.error import HTTPError
from .package import Package, Version
from .errors import DoesNotExists


class PyPI(object):

    blank_packages = False

    def __init__(self, index_url=None):
        self.client = PyPIXmlRpc(
            index_url=index_url or 'https://pypi.python.org/pypi')

    def get(self, name, version=None):
        try:
            name = real_name(name)
        except HTTPError as error:
            if error.code == 404:
                raise DoesNotExists
            else:
                raise error
        package = Package(name, version, client=self.client)
        return package

    @property
    def packages(self):
        """
        TODO: use yield from construction
        """
        packages = self.client.list_packages()
        for name in packages:
            try:
                if not self.blank_packages:
                    package = Package(name)
                else:
                    package = name
                yield package
            except DoesNotExists:
                pass

    def search(self, **kwargs):
        for package in self.client.search(kwargs):
            name, version, summary = package['name'], package['version'], package['summary']
            package = Package(name, version)
            package.summary = summary
            yield package



# coding: utf-8
from setuptools import setup
import pulpitum

setup(name='pulpitum',
      version=pulpitum.__version__,
      author="Sofia Velmer",
      author_email="me@require.pm",
      description='PyPI client for Humans™',
      license="GNU GPLv3",
      keywords="pypi packages",
      url="https://github.com/requirements/pulpitum",
      packages=['pulpitum'],
      package_dir={'pulpitum': 'pulpitum'},
      provides=['pulpitum'],
      classifiers=[
          'Intended Audience :: Developers',
          'Development Status :: 3 - Alpha',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.3',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
      ],
      platforms=['All'],
      install_requires=[
          'pkgtools',
      ],
      )



from distutils.version import StrictVersion, LooseVersion
from pkgtools.pypi import PyPIXmlRpc
from .errors import DoesNotExists


class MetadataStruct(object):

    def __init__(self, data):
        self.__dict__ = data

    def keys(self):
        return self.__dict__.keys()


class Version(object):

    def __init__(self, number):
        number = StrictVersion(number)
        self.number = number

    def __repr__(self):
        return '.'.join(map(str, self.number.version))

    def __str__(self):
        return self.__repr__()

    def version_wrapper(func):
        def decorator(first, second):
            if isinstance(second, Version):
                second = second.number
            return func(first, second)
        return decorator

    @version_wrapper
    def __eq__(self, second):
        return self.number == second

    @version_wrapper
    def __ne__(self, second):
        return self.number != second

    @version_wrapper
    def __lt__(self, second):
        return self.number < second

    @version_wrapper
    def __gt__(self, second):
        return self.number > second

    @version_wrapper
    def __le__(self, second):
        return self.number <= second

    @version_wrapper
    def __ge__(self, second):
        return self.number >= second


class Package(object):

    def __init__(self, name, version=None, client=None):
        self.name = name
        self.client = client or PyPIXmlRpc()
        if version:
            self.version = Version(version)
        else:
            self.version = Version(self.last_release)

    @property
    def metadata(self):
        try:
            metadata = self.cached_metadata
        except AttributeError:
            description = self.client.release_data(self.name, str(self.version))
            if not description:
                raise DoesNotExists
            metadata = self.cached_metadata = MetadataStruct(description)
        return metadata

    @property
    def releases(self):
        try:
            self.__releases = self.__releases
        except AttributeError:
            self.__releases = self.client.package_releases(self.name)
        return sorted(self.__releases)

    @property
    def last_release(self):
        try:
            return self.releases[-1]
        except IndexError:
            raise DoesNotExists

    @property
    def downloads(self):
        urls = self.client.release_urls(self.name, str(self.version))
        return MetadataStruct(urls[0])

    def __getstate__(self):
        name, version = self.name, str(self.version)
        return (name, version)

    def __setstate__(self, state):
        name, version = state
        self.name = name
        self.version = Version(version)

    def __eq__(self, second):
        return self.name == second.name and self.version == second.version

    def __ne__(self, second):
        return self.name != second.name or self.version != second.version

    def __repr__(self):
        return "{}({})".format(self.name, str(self.version))

    def __iter__(self):

        items = (
            ('name', self.name),
            ('version', self.version),
        )

        for item in items:
            yield item



from unittest import TestCase, skip
from pickle import load, dumps
try:
    from StringIO import StringIO
except ImportError:
    from io import BytesIO
from .package import Package, Version
from .api import PyPI
from .errors import DoesNotExists


class VersionTest(TestCase):

    def setUp(self):
        self.pkg_1 = Version('1.4.5')
        self.pkg_2 = Version('1.5.1')

    def test_repr(self):
        self.assertEqual(repr(self.pkg_1), '1.4.5')

    def test_eq(self):
        self.assertFalse(self.pkg_1 == self.pkg_2)

    def test_eq_str(self):
        self.assertFalse(self.pkg_1 == '1.5.1')

    def test_nq(self):
        self.assertTrue(self.pkg_1 != self.pkg_2)

    def test_nq_str(self):
        self.assertTrue(self.pkg_1 != '1.5.1')

    def test_lt(self):
        self.assertTrue(self.pkg_1 < self.pkg_2)

    def test_lt_str(self):
        self.assertTrue(self.pkg_1 < '1.5.1')

    def test_gt(self):
        self.assertFalse(self.pkg_1 > self.pkg_2)

    def test_gt_str(self):
        self.assertFalse(self.pkg_1 > '1.5.1')

    def test_le(self):
        self.assertTrue(self.pkg_1 <= self.pkg_2)

    def test_le_str(self):
        self.assertTrue(self.pkg_1 <= '1.5.1')

    def test_ge(self):
        self.assertTrue(self.pkg_2 >= self.pkg_1)

    def test_ge_str(self):
        self.assertTrue(self.pkg_2 >= '1.4.5')


class PackagesTest(TestCase):

    def setUp(self):
        self.pkg_1 = Package('Django', '1.4.5', client=None)
        self.pkg_2 = Package('Django', '1.5.1', client=None)

    def test_eq(self):
        self.assertFalse(self.pkg_1 == self.pkg_2)
        self.assertTrue(self.pkg_1 == self.pkg_1)

    def test_nq(self):
        self.assertTrue(self.pkg_1 != self.pkg_2)
        self.assertFalse(self.pkg_2 != self.pkg_2)

    def test_dict(self):
        package = Package('Django', '1.5.1')
        result = dict(package)
        self.assertEqual(result['name'], 'Django')
        self.assertEqual(result['version'], '1.5.1')

    def test_releases(self):
        client = PyPI()
        package = client.get('django')
        self.assertTrue(len(package.releases) > 0)

    def test_metadata(self):
        client = PyPI()
        package = client.get('flask')
        metadata = package.metadata
        self.assertEqual(metadata.author, 'Armin Ronacher')

    def test_release_urls(self):
        client = PyPI()
        package = client.get('django')
        info = package.downloads
        self.assertTrue(info.downloads > 0)
        self.assertTrue(info.size > 10)

    def test_repr(self):
        self.assertEqual(repr(self.pkg_1), 'Django(1.4.5)')


class PyPIClientTest(TestCase):

    def setUp(self):
        self.client = PyPI()
        self.pkg_1 = Package('django', '1.4.5')
        self.pkg_2 = Package('django', '1.5.1')

    def test_packages(self):
        packages = self.client.packages
        try:
            package = next(packages)
        except:
            package = packages.next()
        self.assertTrue(package.name)
        self.assertTrue(package.version > '0.0.1')

    def test_blank_packages(self):
        client = PyPI()
        client.blank_packages = True
        packages = client.packages
        try:
            package = next(packages)
        except:
            package = packages.next()
        try:
            self.assertEqual(type(package), str)
        except AssertionError:
            self.assertEqual(type(package), bytes)

    def test_get(self):
        package = self.client.get('django')
        self.assertEqual(package.name, 'Django')
        self.assertTrue(package.version > self.pkg_1.version)
        self.assertTrue(len(package.releases) > 0)

    def test_search(self):
        packages = self.client.search(name='pulpitum')
        self.assertTrue(len(list(packages)) == 1)

    def test_search_with_fetch_releases(self):
        packages = self.client.search(name='pulpitum')
        packages = list(packages)
        self.assertTrue(len(packages) > 0)
        self.assertTrue(type(packages[0].releases), list)

    def test_invalid_pkg(self):
        with self.assertRaises(DoesNotExists):
            package = self.client.get('basjdais')

    def test_pickling(self):
        flask = self.client.get('flask')
        package = dumps(flask)
        try:
            self.assertEqual(type(package), str)
        except AssertionError:
            self.assertEqual(type(package), bytes)
        try:
            flask = load(StringIO(package))
        except NameError:
            flask = load(BytesIO(package))
        self.assertEqual(flask.name, 'Flask')
        self.assertEqual(flask.version, '0.10.1')
